/***
 * GASP: a gaming services platform for mobile multiplayer games.
 * MooDS: a communication protocol used in GASP
 *
 * Copyright (C) 2004 CNAM/INT
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contact: gasp-team@objectweb.org
 *
 * Author: Romain Pellerin
 */
package org.mega.moods.generator;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import org.mega.moods.generator.typesClasses.ElementType;
import org.mega.moods.generator.typesClasses.TableType;
import org.mega.moods.generator.typesClasses.TypeType;

/*
 * Created on 28 juin 2004
 */
// Reste a implmenter la gestion de plusieurs tablo
/**
 * @author Cybernash
 */
public class Generator {
    
    private static int indexOfEntry;
    private static Vector customTypes;

    public static void main(String[] args) {
        String xmlFile = args[0];
        String _package = "";
        boolean isPackaged = false;
        
        if (args.length>=2){
            _package= args[1];
            isPackaged = true;
        }

        try{
            File cTypesFile;
            
            if (isPackaged){
                File dir = new File(_package);
                dir.mkdir();
                cTypesFile = new File(_package+"/CustomTypes.java");
            }
            else cTypesFile = new File("CustomTypes.java");
         
	        cTypesFile.createNewFile();
	        FileWriter cTypesWriter = new FileWriter(cTypesFile);
	        
	        TypesJAXBReader types;
	        String encodeData = "";
	        String encodeMethods ="";
	        String decodeData = "";
	        String decodeMethods = "";
	        
	        //for "compact" mode
	        boolean isCompactMode = false;
	        String finals = "";
	        
	        customTypes = new Vector();
	        
	        try {
	            // get the types instance of the xml file
	            types = new TypesJAXBReader(xmlFile);
	            
	            List listTypes = types.getTypes();
	            String mode = types.getMode();
	            if (mode == null) mode = "";
	            String destClassName = types.getDestClassName();
	            
	            if (mode.equals("compact")) isCompactMode = true;
	            
	            // write package
	            String p = "";
	            if (isPackaged) p = "package "+ _package+";\r\r";
	            
                // write imports
		        String imports = 
		            "import java.io.DataOutputStream;\r" +
		        	"import java.io.DataInputStream;\r" +
		        	"import java.util.Hashtable;\r" +
		        	"import java.util.Enumeration;\r";
		        if (!isCompactMode){
			        for(Iterator iter = listTypes.iterator(); iter.hasNext();){
		                TypeType type = (TypeType) iter.next();
		                imports += "import "+type.getClassName()+";\r";
		                customTypes.addElement(type.getName());
			        }
		        }
			    imports += "\r\r";
			    cTypesWriter.write(p);
			    cTypesWriter.write(imports);
		        

		        
		        // write class declaration
		        cTypesWriter.write("public class CustomTypes{ \r");
		          
		        
		        // start to build root method of the Encoder : encodeData
	            encodeData += "\t public void encodeData(Hashtable h, DataOutputStream dos) throws Exception {\r";
	            encodeData += "\t\t dos.writeByte(h.size());\r";
	            encodeData += "\t\t int nbSend = h.size();\r";
	            encodeData += "\t\t for(int i=0; i<nbSend; i++){\r";
	            encodeData += "\t\t\t Object o  = (Object) h.get(\"\"+i);\r";
    
	            System.out.println("Method void encodeData(Hashtable h, DataOutputStream dos) generated");
	            
	            // start to build root method of the Decoder : decodeData
	            decodeData += "\t public Hashtable decodeData(DataInputStream dis) throws Exception {\r";
	            decodeData += "\t\t Hashtable h = new Hashtable();\r";
	            decodeData += "\t\t byte nb = dis.readByte();\r";          
	            decodeData += "\t\t int nbObj = nb;\r";
	            decodeData += "\t\t while(nb!=0){\r";
	            decodeData += "\t\t\t String key = \"\"+(nbObj - nb);\r"; 
	            decodeData += "\t\t\t byte type = dis.readByte();\r";  
	            
	            System.out.println("Method Hashtable decodeData(DataInputStream dis) generated");
	            
	            int indexType = 6;
	            boolean isFirstIf = true;
	            
	            if (isCompactMode){
	                encodeData += "\t\t\t if (o.getClass().getName().equals(\""+destClassName+"\")){\r";
                    encodeData += "\t\t\t\t dos.writeByte("+indexType+");\r";
                    encodeData += "\t\t\t\t switch ((("+destClassName+") o).getType()){\r";
                    
                    decodeData += "\t\t\t if (type =="+indexType+"){\r";
                    decodeData += "\t\t\t\t "+destClassName+" customType;\r";
                    decodeData += "\t\t\t\t switch (dis.readByte()){\r";
	            }
	   
	            for(Iterator iter = listTypes.iterator(); iter.hasNext();){
	                TypeType type = (TypeType) iter.next();
	                List elements = type.getElement();
	                List tables = type.getTable();
	                indexOfEntry = 0;
	                String typeName = type.getName();
	                
	                // encodeData
	                if (isCompactMode){
	                    int indexFinals = indexType-6;
	                    String typeNameFinal = "TYPE_"+typeName.toUpperCase();
	                    
	                    // build finals
	                    finals += "protected static final int "+typeNameFinal+"="+indexFinals+";\r";
	                    
	                    // build encodeData
	                    encodeData += "\t\t\t\t\t case "+typeNameFinal+":\r"; 
	                    encodeData += "\t\t\t\t\t\t encode"+typeName+"(dos,("+destClassName+") o);\r";
		                encodeData += "\t\t\t\t\t\t break;\r";
	                
		                // build decodeData
		                
		                decodeData += "\t\t\t\t\t case "+typeNameFinal+":\r"; 
		                decodeData += "\t\t\t\t\t\t customType = new "+destClassName+"("+typeNameFinal+");\r";           
		                decodeData += "\t\t\t\t\t\t decode"+typeName+"(dis,customType);\r";
		                decodeData += "\t\t\t\t\t\t h.put(customType,customType);\r";
		                decodeData += "\t\t\t\t\t\t break;\r";
	                }
	                else{
		                // build encodeData
	                    encodeData += "\t\t\t if (o.getClass().getName().equals(\""+type.getClassName()+"\")){\r";
		                encodeData += "\t\t\t\t dos.writeByte("+indexType+");\r";
		                encodeData += "\t\t\t\t encode"+typeName+"(dos,("+typeName+") o);\r";
		                encodeData += "\t\t\t }\r";
		                
		                // build decodeData
		                if (isFirstIf) {isFirstIf = false; decodeData+="\t\t\t ";}
		                else decodeData+="\t\t\t else ";
		                decodeData += "if (type == "+indexType+"){\r";
		                decodeData += "\t\t\t\t "+typeName+" customType = decode"+typeName+"(dis);\r";
		                decodeData += "\t\t\t\t h.put(key,customType);\r";
		                decodeData += "\t\t\t }\r";
	                }
	                	                
	                // write declaration of the new method correponding to the type
	                if (isCompactMode) encodeMethods+="\t public void encode"+typeName+"(DataOutputStream dos,"+destClassName+" customType) throws Exception { \r";
	                else encodeMethods+="\t public void encode"+typeName+"(DataOutputStream dos,"+typeName+" customType) throws Exception { \r";
	                
	                System.out.println("Method void encode"+typeName+"(DataOutputStream dos,"+typeName+" customType) generated");
	                
	                if (isCompactMode) decodeMethods+="\t public void decode"+typeName+"(DataInputStream dis, "+destClassName+" customType) throws Exception { \r";
		            else{
		                decodeMethods+="\t public "+typeName+" decode"+typeName+"(DataInputStream dis) throws Exception { \r";
		                decodeMethods+="\t\t "+typeName+" customType = new "+typeName+"();\r";
		            }
	                System.out.println("Method "+typeName+" decode"+typeName+"(DataInputStream dis) generated");
	                
	                // generate encoder/decoder methods for simple type elements
	                for(Iterator iterElements = elements.iterator(); iterElements.hasNext();){
	                    ElementType element = (ElementType) iterElements.next();
	                    String elementName = element.getName();
	                    elementName = upCaseFirstLetter(elementName);
	                    String elementType = element.getType();
	                    String nameOftype = "";
	                    
	                    if (customTypes.contains(elementType)){ 
	                        encodeMethods+="\t\t encode"+elementType+"(dos,customType.get"+elementName+"());\r";
	                        decodeMethods+="\t\t customType.set"+elementName+"(decode"+elementType+"(dis));\r";
	                    }
	                    else{
	                        nameOftype = getType(elementType);
	                        encodeMethods+=generateEncoding(elementName,nameOftype);
		                    decodeMethods+=generateDecoding(elementName,nameOftype);
	                    }
	                }
	                
	                encodeMethods+="\r";
                    decodeMethods+="\r";
	                
	                int tableNumber = 1;
	                
	                // generate encoder/decoder methods for table types
	                for(Iterator iterTables = tables.iterator(); iterTables.hasNext();){
	                    TableType table = (TableType) iterTables.next();
	                   
	                    String tableName = table.getName();
	                    String tableType = table.getType();
	                    int dimension = table.getDimension();   
	                    String nameOftype = "";
	                    
	                    if (customTypes.contains(tableType)) nameOftype=tableType;
	                    else nameOftype = getType(tableType);
	                    
	                    encodeMethods+=generateGenericTableEncoding(tableNumber, tableName, nameOftype, dimension);
	                    decodeMethods+=generateGenericTableDecoding(tableNumber, tableName,nameOftype, dimension);
	                    encodeMethods+="\r";
	                    decodeMethods+="\r";
	                    
	                    tableNumber++;
	                }
	                
	               encodeMethods+="\t }\r\r";
	               if (!isCompactMode) decodeMethods+="\t\t return customType;\r";
	               decodeMethods+="\t }\r\r";
	               
	               // increment indexType
	               indexType++;
	            }
	            
	            
	            if (isCompactMode){
	                encodeData += "\t\t\t\t\t default: break;\r";
	                encodeData += "\t\t\t\t }\r";
	                encodeData += "\t\t\t }\r";
	                
	                decodeData += "\t\t\t\t\t default: break;\r";
	                decodeData += "\t\t\t\t}\r";
	                decodeData += "\t\t\t\t h.put(key,customType);\r";
	                decodeData += "\t\t\t }\r";             
	            }
	            
	            encodeData += "\t\t}\r";
	            encodeData += "\t}\r";
	            
	            decodeData += "\t\t\t nb--;\r";
	            decodeData += "\t\t}\r";
	            decodeData += "\t\t return h;\r";
	            decodeData += "\t}\r";
	        } 
	        catch (FileNotFoundException e) {
	            e.printStackTrace();
	        }
	        catch (IOException e) {
	            e.printStackTrace();
	        }
	        finally{
	            if (isCompactMode) cTypesWriter.write("\r"+finals+"\r");
	            cTypesWriter.write(encodeMethods);
	            cTypesWriter.write(encodeData);
	            cTypesWriter.write(decodeMethods);
	            cTypesWriter.write(decodeData);
	            cTypesWriter.write("}//CustomTypes");
	            cTypesWriter.flush();
	            cTypesWriter.close();
	            System.out.println(""); 
	            System.out.println("Generation of custom types encoder/decoder finished:"); 
	            System.out.println(" -> see CustomTypes.java in the folder <"+_package+">");
	        }
        }
	    catch (IOException e){
	        e.printStackTrace();
	    }
    }   

    
    /**
     * Return the type of elementType.
     * 
     * @param elementType
     * @return id 
     */
    private static String getType(String elementType) {
        int i = 0;
        if (elementType.toLowerCase().equals("string")) 
            return "String";
        if (elementType.toLowerCase().equals("char")) 
            return "Char";
        if (elementType.toLowerCase().equals("int")) 
            return "Int";
        if (elementType.toLowerCase().equals("long")) 
            return "Long";
        if (elementType.toLowerCase().equals("double")) 
            return "Double";
        if (elementType.toLowerCase().equals("short"))
            return "Short";
        if (elementType.toLowerCase().equals("byte")) 
            return "Byte";
        if (elementType.toLowerCase().equals("boolean"))
            return "Boolean";
        else return "";
    } 
    
    
    /**
     * Generate the encoding of an element.
     * 
     * @param type the id of type
     * @return string to write in Encoder class
     */
    private static String generateEncoding(String elementName,String type) {
        String toEncoder = "";
        if (type.equals("String")) type = "UTF";
        toEncoder="\t\t dos.write"+type+"(customType.get"+elementName+"());\r";
        return toEncoder;
    }
    
    
    /**
     * Generate the decoding of an element.
     * 
     * @param elementName the name of the element
     * @param type the id of element type
     * @return string to write in Decoder class
     */
    private static String generateDecoding(String elementName, String type) {
        String toDecoder = "\t\t customType.set"+elementName+"(dis.read";
        if (type.equals("String")) toDecoder+="UTF";
        else toDecoder+=type;
        toDecoder += "()); \r";
        return toDecoder;
    } 
    
    
    /**
     * Generate the decoding code for a table.
     * 
     * @param table
     * @param tableName
     * @param tableType
     */
    private static String generateTableEncoding(String tableName, String type, int dimension) {
        String toEncoder="";
       
        if (type.equals("String")) toEncoder+="\t\t "+type;
        else toEncoder+="\t\t "+type.toLowerCase();
        toEncoder+=getDimension(dimension)+" table = customType.get"+upCaseFirstLetter(tableName)+"();\r";;
        toEncoder+="\t\t dos.writeByte(table.length);\r";
        toEncoder+="\t\t for(int i=0; i<"+dimension+"; i++){\r";
        toEncoder+="\t\t\t for(int j=0; i<table.length; i++){\r";
        toEncoder+="\t\t\t\t dos.write";
        if (type.equals("String")) toEncoder+="UTF";
        else toEncoder+=type;
        toEncoder+="(table[i][j]);\r";
        toEncoder+="\t\t\t }\r";
        toEncoder+="\t\t }\r";
        
        return toEncoder;        
    }
    
    
    /**
     * Generate the decoding code for a table.
     * 
     * @param n table number n
     * @param table
     * @param tableName
     * @param tableType
     */
    private static String generateGenericTableEncoding(int n, String tableName, String type, int dimension) {
        String t = type.toLowerCase();
        String toEncoder="";
        String nTable= "table"+n;
        String nTableL= "table"+n+"_l";
        
        boolean noLowerCase = false;
        if (type.equals("String")||customTypes.contains(type)) noLowerCase = true;
        
        if (noLowerCase) toEncoder+="\t\t "+type;
        else toEncoder+="\t\t "+t;
        toEncoder+=getDimension(dimension)+" "+nTable +"= customType.get"+upCaseFirstLetter(tableName)+"();\r";
        
        int d = 0;
        
        while(d<dimension){
            String ind = indent(d);
            toEncoder+="\t\t"+ind+" int "+nTableL+d+" = "+nTable+""+tableLength(d)+".length;\r";
            toEncoder+="\t\t"+ind+" dos.writeShort((short) "+nTableL+d+");\r";
            toEncoder+="\t\t"+ind+" for(int i"+d+"=0; i"+d+"<"+nTableL+d+"; i"+d+"++){\r";
            d++;
        }
        if (customTypes.contains(type)){
            toEncoder+="\t\t"+indent(dimension)+" encode"+type+"(dos,"+nTable+accessTable(dimension)+");\r";
        }
        else{
	        toEncoder+="\t\t"+indent(dimension)+" dos.write";
	        if (type.equals("String")) toEncoder+="UTF";
	        else toEncoder+=type;
	        toEncoder+="("+nTable+accessTable(dimension)+");\r";
        }
        
        toEncoder+=closeQuote(dimension);

        return toEncoder;        
    }
    
    
    /**
     * Generate the decoding code for a table.
     * 
     * @param table
     * @param tableName
     * @param tableType
     */
    private static String generateGenericTableDecoding(int n, String tableName, String type, int dimension) {
        String t = type.toLowerCase();
        String toDecoder="";
        String nTable= "table"+n;
        String nTableL= "table"+n+"_l";

        boolean noLowerCase = false;
        if (type.equals("String")||customTypes.contains(type)) noLowerCase = true;
        
        if (noLowerCase) toDecoder+="\t\t "+type;
        else toDecoder+="\t\t "+t;
        toDecoder+=getDimension(dimension)+" "+nTable+";\r";
        
        int d = 0;
        
        while(d<dimension){
            String ind = indent(d);
            toDecoder+="\t\t"+ind+" short "+nTableL+d+" = dis.readShort();\r";
            toDecoder+="\t\t"+ind+" "+nTable;
            if (noLowerCase) toDecoder+=initTable(nTableL,dimension,d,type);
            else toDecoder+=initTable(nTableL,dimension,d,t);
            toDecoder+=";\r";
            toDecoder+="\t\t"+ind+" for(int i"+d+"=0; i"+d+"<"+nTableL+d+"; i"+d+"++){\r";
            d++;
        }
        
        if (customTypes.contains(type)){
            toDecoder+="\t\t"+indent(dimension)+" "+nTable+accessTable(dimension)+"= decode"+type+"(dis);\r";
        }
        else{
            toDecoder+="\t\t"+indent(dimension)+" "+nTable+accessTable(dimension)+"= dis.read";
        	if (type.equals("String")) toDecoder+="UTF";
        	else toDecoder+=type;
        	toDecoder+="();\r";
        }
        
        toDecoder+= closeQuote(dimension);
        toDecoder+="\t\t customType.set"+upCaseFirstLetter(tableName)+"("+nTable+");\r";

        return toDecoder;        
    }
    
    /**
     * @param dimension
     * @param d
     * @return
     */
    private static String initTable(String varName, int dimension, int d, String type) {
        String dec = "";
        int n=0;
        while(n<d){
            dec+="[i"+n+"]";
            n++;
        }
        dec+=" = new "+type+"["+varName+n+"]";
        n++;
        while(n<dimension){ dec+="[]"; n++;}
        return dec;
    }


    /**
     * Generate string : ("}")*
     * 
     * @param dimension
     */
    private static String closeQuote(int d) {
        int i = 0;
        String ind = "";
        while (i<d){
            ind+=indent(d-i+1)+"}\r";
            i++;
        }
        return ind;
    }


    /**
     * Generate string : ("[ii]")*
     * 
     * @param d dimension
     * @return
     */
    private static String accessTable(int d) {
        int i = 0;
        String form = "";
        while (i<d){
            form+="[i"+i+"]";
            i++;
        }
        return form;
    }


    /**
     * Generate string : "table"("[]")*
     * 
     * @param d dimension
     * @return
     */
    private static String tableLength(int d) {
        int i = 0;
        int j;
        String form = "";
        while (i<d){
            form+="[i"+i+"]";
            i++;
        }
        return form;
    }


    /**
     * Generate string : ("\t")*
     * 
     * @param d dimension
     * @return
     */
    private static String indent(int d) {
        int i = 0;
        String ind = "";
        while (i<d){
            ind+="\t";
            i++;
        }
        return ind;
    }


    /**
     * Generate the encoding code for a table.
     * 
     * @param table
     * @param tableName
     * @param tableType
     */
    private static String generateTableDecoding(String tableName, String type, int dimension) {
        String toDecoder="";
  
         if (type.equals("String")) toDecoder+="\t\t "+type;
         else toDecoder+="\t\t "+type.toLowerCase()+getDimension(dimension)+" table";
         toDecoder+=" = new ";
         if (type.equals("String")) toDecoder+=type;
         else toDecoder+=type.toLowerCase();
 
         toDecoder+=setDimension(dimension);
         toDecoder+=";\r";
         toDecoder+="\t\t for(int i=0; i<"+dimension+"; i++){\r";
         toDecoder+="\t\t\t for(int j=0; i<table.length; i++){\r";
         toDecoder+="\t\t\t\t table[i][j] = dis.read";
         if (type.equals("String")) toDecoder+="UTF";
         else toDecoder+=type;
         toDecoder+="();\r";
         toDecoder+="\t\t\t }\r";
         toDecoder+="\t\t }\r";
         toDecoder+="\t\t customType.set"+upCaseFirstLetter(tableName)+"(table);\r";
         
         return toDecoder;   
    }


    /**
     * Generate table dimension code.
     * 
     * @param dimension
     * @return the string of the dimension code
     */
    private static String getDimension(int dimension) {
        String dim="";
        
        while(dimension>0){ 
            dim+="[]";
            dimension--;
        }
        return dim;
    }
    
    
    /**
     * @param dimension
     * @return
     */
    private static String setDimension(int dimension) {
        String dim="";
        if (dimension>0){
            dim+="["+dimension+"]";
            
        }
        dim+="[dis.readByte()]";
        return dim;
    }
    
    
    /**
     * @param s
     * @return
     */
    private static String upCaseFirstLetter(String s) {
        char[] cTab = s.toCharArray();
        cTab[0] = (cTab[0]+"").toUpperCase().charAt(0);
        return new String(cTab);
    }
    
    
}
